﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.SqlClient;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;

namespace DevLeap.Linq.LinqToDataSet {
    using NorthwindTableAdapters;

    public partial class DemoLinqToDataSet : Form {
        const string ConnectionString = "Database=Northwind;Trusted_Connection=yes";

        public DemoLinqToDataSet() {
            InitializeComponent();
        }

        public static DataSet LoadDataSetUsingDataAdapter() {
            //
            // Listing 11-1
            //
            const string QueryOrders = @"
SELECT  OrderID, OrderDate, Freight, ShipName, 
        ShipAddress, ShipCity, ShipCountry
FROM    Orders
WHERE   CustomerID = @CustomerID

SELECT     od.OrderID, od.UnitPrice, od.Quantity, CAST( NULLIF( od.Discount, 0 ) AS FLOAT ) AS Discount,
           p.[ProductName]
FROM       [Order Details] od
INNER JOIN Orders o 
      ON   o.[OrderID] = od.[OrderID]
LEFT JOIN  Products p
     ON    p.[ProductID] = od.[ProductID]
WHERE      o.CustomerID = @CustomerID";

            DataSet ds = new DataSet("CustomerOrders");
            SqlDataAdapter da = new SqlDataAdapter(QueryOrders, ConnectionString);
            da.SelectCommand.Parameters.AddWithValue("@CustomerID", "QUICK");
            da.TableMappings.Add("Table", "Orders");
            da.TableMappings.Add("Table1", "OrderDetails");
            da.Fill(ds);

            return ds;
        }

        public static void ShowDataAdapter() {
            DataSet ds = LoadDataSetUsingDataAdapter();

            CreateViewer(ds.Tables["Orders"]);

            CreateViewer(ds.Tables["OrderDetails"]);
        }

        public static void LoadDataSetWithLinq() {
            //
            // Listing 11-3
            //
            NorthwindDataContext db = new NorthwindDataContext(ConnectionString);
            string filterCustomerID = "QUICK";

            var customerOrders =
                from o in db.Orders
                where o.CustomerID == filterCustomerID
                select new {
                    o.OrderID,
                    o.OrderDate,
                    o.Freight,
                    o.ShipName,
                    o.ShipAddress,
                    o.ShipCity,
                    o.ShipCountry
                };

            var customerOrderDetails =
                from od in db.Order_Details
                join o in db.Orders
                       on od.OrderID equals o.OrderID
                join p in db.Products
                       on od.ProductID equals p.ProductID
                where o.CustomerID == filterCustomerID
                select new {
                    od.UnitPrice,
                    od.Quantity,
                    od.Discount,
                    p.ProductName
                };

            DataSet ds = new DataSet("CustomerOrders");
            ds.Tables.Add(customerOrders.CreateDataTable("Orders"));
            ds.Tables.Add(customerOrderDetails.CreateDataTable("OrderDetails"));

            CreateViewer(ds.Tables["Orders"]);
            //CreateViewer(ds.Tables["OrderDetails"]);
        }

        public static void LoadDataSetWithCopyToDataTable() {
            //
            // Listing 11-4
            //
            DataSet ds = LoadDataSetUsingDataAdapter();

            var highDiscountOrders =
                from o in ds.Tables["OrderDetails"].AsEnumerable()
                where o.Field<double?>("Discount") > 0.2
                select o;

            DataTable highDiscountOrdersTable = highDiscountOrders.CopyToDataTable();
            highDiscountOrdersTable.TableName = "HighDiscountOrders";
            ds.Tables.Add(highDiscountOrdersTable);

            CreateViewer(highDiscountOrdersTable);
        }
        
        public static void AddDataWithCopyToDataTable() {
            //
            // Listing 11-5
            //
            DataSet ds = LoadDataSetUsingDataAdapter();

            var highDiscountOrders =
                from o in ds.Tables["OrderDetails"].AsEnumerable()
                where o.Field<double?>("Discount") > 0.2
                select o;

            DataTable selectionTable = highDiscountOrders.CopyToDataTable();

            var lowDiscountOrders =
                from o in ds.Tables["OrderDetails"].AsEnumerable()
                where o.Field<double?>("Discount") < 0.05
                select o;

            lowDiscountOrders.CopyToDataTable(selectionTable, LoadOption.PreserveChanges);

            CreateViewer(selectionTable);
        }

        private static void CreateViewer(object source) {
            Viewer viewer = new Viewer();
            viewer.dataGridView.DataSource = source;
            viewer.Show();
        }

        public static void QueryDataTable() {
            //
            // Listing 11-6
            //
            DataSet ds = LoadDataSetUsingDataAdapter();
            DataTable orders = ds.Tables["Orders"];

            var query =
                from o in orders.AsEnumerable()
                where o.Field<DateTime>("OrderDate").Year >= 1998
                orderby o.Field<DateTime>("OrderDate") descending
                select o;

            CreateViewer(query.CopyToDataTable());
        }

        public static void QueryJoinDataTables() {
            //
            // Listing 11-7
            //
            DataSet ds = LoadDataSetUsingDataAdapter();
            DataTable orders = ds.Tables["Orders"];
            DataTable orderDetails = ds.Tables["OrderDetails"];

            var query =
                from o in orders.AsEnumerable()
                join od in orderDetails.AsEnumerable()
                        on o.Field<int>("OrderID") equals od.Field<int>("OrderID")
                        into orderLines
                where o.Field<DateTime>("OrderDate").Year >= 1998
                orderby o.Field<DateTime>("OrderDate") descending
                select new
                {
                    OrderID = o.Field<int>("OrderID"),
                    OrderDate = o.Field<DateTime>("OrderDate"),
                    Amount = orderLines.Sum(
                                 od => od.Field<decimal>("UnitPrice")
                                       * od.Field<short>("Quantity"))
                };

            CreateViewer(query.CreateDataTable("Result"));
        }

        public static void QueryUsingImplicitJoin() {
            //
            // Listing 11-8
            //
            DataSet ds = LoadDataSetUsingDataAdapter();
            DataTable orders = ds.Tables["Orders"];
            DataTable orderDetails = ds.Tables["OrderDetails"];
            ds.Relations.Add("OrderDetails",
                              orders.Columns["OrderID"],
                              orderDetails.Columns["OrderID"]);

            var query =
                from o in orders.AsEnumerable()
                where o.Field<DateTime>("OrderDate").Year >= 1998
                orderby o.Field<DateTime>("OrderDate") descending
                select new
                {
                    OrderID = o.Field<int>("OrderID"),
                    OrderDate = o.Field<DateTime>("OrderDate"),
                    Amount = o.GetChildRows("OrderDetails").Sum(
                                 od => od.Field<decimal>("UnitPrice")
                                       * od.Field<short>("Quantity"))
                };

            CreateViewer(query.CreateDataTable("Result"));
        }
        public static void CreateDataView() {
            DataSet ds = LoadDataSetUsingDataAdapter();
            DataTable orders = ds.Tables["Orders"];

            //
            // Listing 11-9
            //
            string filter = String.Format("OrderDate >= #{0:d}#", new DateTime(1998, 1, 1));
            string sort = "OrderDate DESC";
            DataView legacyView = new DataView(orders, filter, sort, DataViewRowState.CurrentRows);

            CreateViewer(legacyView);

            //
            // Listing 11-10
            //
            var query =
                from o in orders.AsEnumerable()
                where o.Field<DateTime>("OrderDate").Year >= 1998
                orderby o.Field<DateTime>("OrderDate") descending
                select o;

            DataView view = query.AsDataView();
            CreateViewer(view);

            //
            // After executing the following line, you will replace the Where condition of the LINQ query
            // but you still will keep the behavior of the orderby condition
            view.RowFilter = String.Format("OrderDate < #{0:d}#", new DateTime(1998, 1, 1));
            view.Sort = null;
        }

        public static void QueryTypedDataSet() {
            Northwind ds = new Northwind();
            string filterCustomerID = "QUICK";
           
            OrdersTableAdapter ordersTableAdapter = new OrdersTableAdapter();
            Order_DetailsTableAdapter orderDetailsTableAdapter = new Order_DetailsTableAdapter();
            ordersTableAdapter.FillByCustomer(ds.Orders, filterCustomerID);
            orderDetailsTableAdapter.FillByCustomer(ds.Order_Details, filterCustomerID);

            //
            // Listing 11-11
            //
            var query =
                from o in ds.Orders
                where o.OrderDate.Year >= 1998
                orderby o.OrderDate descending
                select new
                {
                    o.OrderID,
                    o.OrderDate,
                    Amount = o.GetOrder_DetailsRows().Sum(
                                 od => od.UnitPrice * od.Quantity)
                };

            CreateViewer(query.ToList());
        }

        public void UntypedNullCheck() {
            DataSet ds = LoadDataSetUsingDataAdapter();
            DataTable orderDetails = ds.Tables["OrderDetails"];

            //
            // Listing 11-12
            //
            foreach (DataRow r in orderDetails.Rows) {
                double? discount = r.Field<double?>("Discount");
                if (discount.HasValue && discount.Value > 0.10) {
                    r.SetField<double?>("Discount", 0.10);
                }
            }

            CreateViewer(orderDetails);
        }

        public void DemoDataRowComparer() {
            Northwind ds = new Northwind();
            string filterCustomerID = "QUICK";
           
            OrdersTableAdapter ordersTableAdapter = new OrdersTableAdapter();
            ordersTableAdapter.FillByCustomer(ds.Orders, filterCustomerID);

            //
            // Listing 11-13
            //
            var queryOldOrders =
                from o in ds.Orders
                where o.OrderDate.Year <= 1996
                select o;

            var queryLowFreightOrders =
                from o in ds.Orders
                where o.Freight < 50
                select o;

            DataTable oldOrders = queryOldOrders.CopyToDataTable();
            DataTable lowFreightOrders = queryLowFreightOrders.CopyToDataTable();

            // Find the intersection of the two queries
            var queryOrders = oldOrders.AsEnumerable().Intersect(
                                  lowFreightOrders.AsEnumerable(),
                                  DataRowComparer.Default);

            DataTable orders = queryOrders.CopyToDataTable();
            //CreateViewer(ds.Orders);
            //CreateViewer(oldOrders);
            //CreateViewer(lowFreightOrders);
            CreateViewer(orders);
            
        }

        private void listing71DataAdapterToolStripMenuItem_Click(object sender, EventArgs e) {
            ShowDataAdapter();
        }

        private void listing73CreateDataTableToolStripMenuItem_Click(object sender, EventArgs e) {
            LoadDataSetWithLinq();
        }

        private void listing74CopyToDataTableToolStripMenuItem_Click(object sender, EventArgs e) {
            LoadDataSetWithCopyToDataTable();
        }

        private void listing76QueryToolStripMenuItem_Click(object sender, EventArgs e) {
            QueryDataTable();
        }

        private void listing77SimpleJoinToolStripMenuItem_Click(object sender, EventArgs e) {
            QueryJoinDataTables();
        }

        private void listing78ImplicitJoinToolStripMenuItem_Click(object sender, EventArgs e) {
            QueryUsingImplicitJoin();
        }

        private void listing75AddRowsWithCopyToDataTableToolStripMenuItem_Click(object sender, EventArgs e) {
            AddDataWithCopyToDataTable();
        }

        private void listing79_10DataViewToolStripMenuItem_Click(object sender, EventArgs e) {
            CreateDataView();
        }

        private void listing711QueryTypedDatasetToolStripMenuItem_Click(object sender, EventArgs e) {
            QueryTypedDataSet();
        }

        private void listing712UntypedFieldNullCheckToolStripMenuItem_Click(object sender, EventArgs e) {
            UntypedNullCheck();
        }

        private void listing713DataRowComparerToolStripMenuItem_Click(object sender, EventArgs e) {
            DemoDataRowComparer();
        }



    }
}
