﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Linq;
using System.Data.Linq.Mapping;
using System.Linq.Expressions;
using System.Diagnostics;
using System.Data;
using System.Reflection;

namespace DevLeap.Linq.LinqToSql.SimpleQuery {
    public class Marco {
        public string x;
    }
    //
    // Listing 5-1
    //
    [Table(Name = "Customers")]
    public class Customer {
        [Column(IsPrimaryKey = true)]
        public string CustomerID;
        [Column]
        public string CompanyName;
        [Column]
        public string City;
        [Column(Name = "Region")]
        public string State;
        [Column]
        public string Country;
    }

    //
    // Listing 5-3
    //
    [Table(Name = "Contacts")]
    [InheritanceMapping(Code = "Customer", Type = typeof(CustomerContact))]
    [InheritanceMapping(Code = "Supplier", Type = typeof(SupplierContact))]
    [InheritanceMapping(Code = "Shipper", Type = typeof(ShipperContact))]
    [InheritanceMapping(Code = "Employee", Type = typeof(Contact), IsDefault = true)]
    public class Contact {
        [Column(IsPrimaryKey = true)]
        public int ContactID;
        [Column(Name = "ContactName")]
        public string Name;
        [Column]
        public string Phone;
        [Column(IsDiscriminator = true)]
        public string ContactType;
    }

    public class CompanyContact : Contact {
        [Column(Name = "CompanyName")]
        public string Company;
    }
    public class CustomerContact : CompanyContact {
    }

    public class SupplierContact : CompanyContact {
    }

    public class ShipperContact : CompanyContact {
        public string Shipper {
            get { return Company; }
            set { Company = value; }
        }
    }

    [Table(Name = "Categories")]
    public class Category {
        [Column(IsPrimaryKey = true)]
        public int CategoryID;
        [Column]
        public string CategoryName;
    }

    [Table(Name = "Orders")]
    public class Order {
        [Column(IsPrimaryKey = true)]
        public int OrderID;
        [Column]
        public string CustomerID;
        [Column]
        public decimal Freight;
    }

    class QueryDemo {
        // 
        // Listing 5-2
        // 
        public static void SimpleQuery() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Customer> Customers = db.GetTable<Customer>();

            var query =
                from c in Customers
                where c.Country == "USA"
                        && c.State == "WA"
                select new { c.CustomerID, c.CompanyName, c.City };

            foreach (var row in query) {
                Console.WriteLine(row);
           }
        
            Console.WriteLine(query);
            //Console.WriteLine(db.GetCommand(query).CommandText);

        }

        public static void SimpleQueryCe() {
            DataContext db = new DataContext(Connections.ConnectionStringCompact);
            Table<Customer> Customers = db.GetTable<Customer>();

            var query =
                from c in Customers
                where c.Country == "USA"
                        && c.State == "WA"
                select new { c.CustomerID, c.CompanyName, c.City };

            foreach (var row in query) {
                Console.WriteLine(row);
            }

            Console.WriteLine(query);

        }

        public static void Inheritance() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Contact> contacts = db.GetTable<Contact>();
            db.Log = Console.Out;

            // 
            // Listing 5-4
            // 
            var queryTyped =
                from c in contacts.OfType<CustomerContact>()
                select c;

            var queryFiltered =
                from c in contacts
                where c is CustomerContact
                select c;

            foreach (var row in queryTyped) {
                Console.WriteLine(row.Company);
            }

            // We need an explicit cast to access the CostumerContact members
            foreach (CustomerContact row in queryFiltered) {
                Console.WriteLine(row.Company);
            }
        }

        public static void UniqueIdentity() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Contact> contacts = db.GetTable<Contact>();
            db.Log = Console.Out;

            // 
            // Listing 5-5
            // 
            var queryTyped =
                from c in contacts.OfType<CustomerContact>()
                orderby c.ContactID
                select c;

            var queryFiltered =
                from c in contacts
                where c is CustomerContact
                orderby c.ContactID
                select c;

            Contact c1 = null;
            Contact c2 = null;
            foreach (var row in queryTyped.Take(1)) {
                c1 = row;
            }
            foreach (var row in queryFiltered.Take(1)) {
                c2 = row;
            }
            Contact c3 = new Contact();
            c3.ContactID = c1.ContactID;
            c3.ContactType = c1.ContactType;
            c3.Name = c1.Name;
            c3.Phone = c1.Phone;
            Debug.Assert(c1 == c2); // same instance
            Debug.Assert(c1 != c3); // different objects
        }

        // 
        // Listing 5-11
        // 
        public static void QueryManipulation() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Customer> Customers = db.GetTable<Customer>();
            db.Log = Console.Out;

            // All Customers
            var query =
                from c in Customers
                select new { c.CompanyName, c.State, c.Country };

            DisplayTop(query, 10);

            // User interaction add a filter 
            // to the previous query 
            // Customers from USA
            query =
                from c in query
                where c.Country == "USA"
                select c;

            DisplayTop(query, 10);

            // User interaction add another 
            // filter to the previous query 
            // Customers from WA, USA
            query =
                from c in query
                where c.State == "WA"
                select c;

            DisplayTop(query, 10);
        }

        static void DisplayTop<T>(IQueryable<T> query, int rows) {
            foreach (var row in query.Take(rows)) {
                Console.WriteLine(row);
            }
        }

        public static void QuerySingle() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Customer> Customers = db.GetTable<Customer>();
            db.Log = Console.Out;

            // Retrieve a single customer
            var query =
                from c in Customers
                where c.CustomerID == "ANATR"
                select c;

            using (var enumerator = query.GetEnumerator()) {
                if (enumerator.MoveNext()) {
                    var customer = enumerator.Current;
                    Console.WriteLine("{0} {1}", customer.CustomerID, customer.CompanyName);
                }
            }
            // Simpler way to retrieve a single customer
            {
                var customer = Customers.Single(c => c.CustomerID == "ANATR");
                Console.WriteLine("{0} {1}", customer.CustomerID, customer.CompanyName);

                // This call will read the data from the database again
                db.Refresh(RefreshMode.OverwriteCurrentValues, customer);
            }

        }

        static public void Projection() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Customer> Customers = db.GetTable<Customer>();
            db.Log = Console.Out;

            var query =
                from c in Customers
                where c.Country == "USA"
                select new { c.CustomerID, Name = c.CompanyName.ToUpper() } into r
                orderby r.Name
                select r;

            foreach (var company in query) {
                Console.WriteLine(company);
            }


            var queryBad =
                from c in Customers
                where c.Country == "USA"
                select new CustomerData(c.CustomerID, c.CompanyName.ToUpper()) into r
                orderby r.Name
                select r;
            // The execution of query2 would throw an exception
            //foreach( var company in queryBad ) {
            //    Console.WriteLine( company );
            //}

            var queryParamConstructor =
                from c in Customers
                where c.Country == "USA"
                orderby c.CompanyName
                select new CustomerData(c.CustomerID, c.CompanyName.ToUpper());

            foreach (var company in queryParamConstructor) {
                Console.WriteLine(company);
            }

            var queryGood =
                from c in Customers
                where c.Country == "USA"
                select new CustomerData { CustomerID = c.CustomerID, Name = c.CompanyName.ToUpper() } into r
                orderby r.Name
                select r;
            
            foreach (var company in queryGood) {
                Console.WriteLine(company);
            }

        }

        public static void CallStoredProcedure() {
            SampleDb db = new SampleDb(Connections.ConnectionString);
            db.Log = Console.Out;
            
            foreach (var row in db.CustomersByCity("London")) {
                Console.WriteLine("{0} {1}", row.CustomerID, row.CompanyName);
            } 

            IMultipleResults results = db.TwoCustomerGroups();
            foreach (var row in results.GetResult<CustomerInfo>()) {
                Console.WriteLine("{0} {1} - {2} ({3})", row.CustomerID, row.City, row.CompanyName, row.ContactName);
            }
            Console.WriteLine("---");
            foreach (var row in results.GetResult<CustomerShortInfo>()) {
                Console.WriteLine("{0} {1} - {2}", row.CustomerID, row.City, row.CompanyName);
            }

            // ******************************
            // DEMO COMPOSABILITY STORED PROCEDURE (does not compose a SQL query!)
            //
            var query =
                from c in db.CustomersByCity("London")
                where c.CompanyName.Length > 15
                select new { c.CustomerID, c.CompanyName };

            Console.WriteLine("** COMPOSABILITY **");
            foreach (var row in query) {
                Console.WriteLine(row);
            }
            //
            // ******************************


        }

        public static void UserDefinedFunctions() {
            SampleDb db = new SampleDb(Connections.ConnectionString);
            Table<Category> Categories = db.GetTable<Category>();
            db.Log = Console.Out;

            var query =
                from c in Categories
                select new {
                    c.CategoryID,
                    c.CategoryName,
                    MinPrice = db.MinUnitPriceByCategory(c.CategoryID)
                };

            foreach (var row in query) {
                Console.WriteLine(row);
            }

            Table<Order> Orders = db.GetTable<Order>();
            var queryCustomers =
                from c in db.CustomersByCountry("USA")
                join o in Orders
                       on c.CustomerID equals o.CustomerID
                       into orders
                select new { c.CustomerID, c.CompanyName, orders };

            foreach (var row in queryCustomers) {
                Console.WriteLine(row);
            }
        }

        //
        // Listing 5-16
        //
        public static void CompiledQueriesLocal() {
            DataContext db = new DataContext(Connections.ConnectionString);
            Table<Customer> Customers = db.GetTable<Customer>();

            var query =
                CompiledQuery.Compile(
                    (DataContext context, string filterCountry) =>
                        from c in Customers
                        where c.Country == filterCountry
                        select new { c.CustomerID, c.CompanyName, c.City });

            db.Log = Console.Out;

            Console.WriteLine("===================");
            foreach (var row in query(db, "USA")) {
                Console.WriteLine(row);
            }
            Console.WriteLine("===================");
            foreach (var row in query(db, "Italy")) {
                Console.WriteLine(row);
            }
        }

        //
        // Listing 5-17
        //
        static Func<SampleDb, string, IQueryable<Customer>>
            CustomerByCountry =
                CompiledQuery.Compile(
                    (SampleDb db, string filterCountry) =>
                        from c in db.Customers
                        where c.Country == filterCountry
                        select c);

        public static void CompiledQueriesStatic() {
            SampleDb db = new SampleDb(Connections.ConnectionString);

            db.Log = Console.Out;

            Console.WriteLine("==========USA===========");
            foreach (var row in CustomerByCountry(db, "USA")) {
                Console.WriteLine(row.CustomerID);
            }
            Console.WriteLine("=========ITALY==========");
            foreach (var row in CustomerByCountry(db, "Italy")) {
                Console.WriteLine(row.CustomerID);
            }
        }


        public static void GraphTraversal() {
            nwDataContext db = new nwDataContext(Connections.ConnectionString);
            db.Log = Console.Out;

            //
            // Listing 5-18
            //
            var queryJoin =
                from p in db.Products
                join o in db.Order_Details
                       on p.ProductID equals o.ProductID
                       into OrdersProduct
                where p.ProductName == "Chocolade"
                select OrdersProduct.Sum(o => o.Quantity);
            var quantityJoin = queryJoin.Single();
            Console.WriteLine(quantityJoin);

            //
            // Listing 5-19
            //
            var queryAssociation =
                from p in db.Products
                where p.ProductName == "Chocolade"
                select p.Order_Details.Sum(o => o.Quantity);
            var quantityAssociation = queryAssociation.Single();
            Console.WriteLine(quantityAssociation);

            //
            // Listing 5-20
            //
            var chocolade = db.Products.Single(p => p.ProductName == "Chocolade");
            var quantityValue = chocolade.Order_Details.Sum(o => o.Quantity);
            Console.WriteLine(quantityValue);

            var repeatCalc = chocolade.Order_Details.Sum(o => o.Quantity);
            Console.WriteLine(repeatCalc);

            //
            // Listing 5-21
            //
            var quantityChocolade = db.Products.Single(p => p.ProductName == "Chocolade")
                                    .Order_Details.Sum(o => o.Quantity);
            Console.WriteLine(quantityChocolade);
        }


        //
        // Listing 5-22
        //
        public static void DirectQuery() {
            SampleDb db = new SampleDb(Connections.ConnectionString);
            db.Log = Console.Out;

            var query = db.ExecuteQuery<EmployeeInfo>(@"
WITH EmployeeHierarchy (EmployeeID, LastName, FirstName, 
     ReportsTo, HierarchyLevel) AS
(  SELECT EmployeeID,LastName, FirstName, 
          ReportsTo, 1 as HierarchyLevel
   FROM   Employees
   WHERE  ReportsTo IS NULL

   UNION ALL

   SELECT      e.EmployeeID, e.LastName, e.FirstName, 
               e.ReportsTo,eh.HierarchyLevel + 1 AS HierarchyLevel
   FROM        Employees e
   INNER JOIN  EmployeeHierarchy eh 
           ON  e.ReportsTo = eh.EmployeeID 
)
SELECT *
FROM EmployeeHierarchy
ORDER BY HierarchyLevel, LastName, FirstName");
            foreach (var row in query) {
                Console.WriteLine(row);
            }
        }

        public class CompanyOrders {
            public string CompanyName;
            public DateTime FirstOrderDate;
            public DateTime LastOrderDate;

            public override string ToString() {
                return String.Format(
                    "Customer={0} FirstOrder={1} LastOrder={2}", 
                    CompanyName, FirstOrderDate, LastOrderDate);
            }
        }
        //
        // Listing 5-23
        //
        public static void DirectQueryParametric() {
            SampleDb db = new SampleDb(Connections.ConnectionString);
            db.Log = Console.Out;

            var query = db.ExecuteQuery<CompanyOrders>(@"
            SELECT 
                c.CompanyName, 
                MIN( o.OrderDate ) AS FirstOrderDate,
                MAX( o.OrderDate ) AS LastOrderDate
            FROM Customers c
            LEFT JOIN Orders o
                   ON o.CustomerID = c.CustomerID
            GROUP BY c.CustomerID, c.CompanyName
            HAVING COUNT(o.OrderDate) > 0
               AND MIN( o.OrderDate ) BETWEEN {0} AND {1}
            ORDER BY FirstOrderDate ASC",
                new DateTime(1997, 1, 1),
                new DateTime(1997, 12, 31));

            foreach (var row in query) {
                Console.WriteLine(row);
            }
        }

        //
        // Listing 5-27
        //
        public static void DemoContains() {
            nwDataContext db = new nwDataContext(Connections.ConnectionString);

            var query =
                from c in db.Customers
                where !(from o in db.Orders
                        select o.CustomerID)
                       .Contains(c.CustomerID)
                select new { c.CustomerID, c.CompanyName };

            Console.WriteLine(query);

            foreach (var c in query) {
                Console.WriteLine(c);
            }
        }

        //
        // Listing 5-28
        //
        public static void DemoContainsConstants() {
            nwDataContext db = new nwDataContext(Connections.ConnectionString);

            //var cities = new string[] { "London", "Seattle" };
            //var query =
            //    from c in db.Customers
            //    where cities.Contains(c.City)
            //    select new { c.CustomerID, c.CompanyName, c.City };

            var query =
                from c in db.Customers
                where (new string[] { "London", "Seattle" }).Contains(c.City)
                select new { c.CustomerID, c.CompanyName, c.City };

            Console.WriteLine(query);

            foreach (var c in query) {
                Console.WriteLine(c);
            }
        }
        public static void QueryReduction() {
            nwDataContext db = new nwDataContext(Connections.ConnectionString);
            // 
            // Listing 5-29
            //
            var query = 
                from s in ( 
                    from c in db.Customers
                    where c.CompanyName.Length > 10
                    select new { c.CustomerID, c.CompanyName, c.ContactName, c.City,
                                 c.Country, c.ContactTitle, c.Address }
                )
                where s.Country == "UK"
                select new { s.CustomerID, s.CompanyName, s.City };

            Console.WriteLine( query );
        }

        public static void AggregationNulls() {
            DataContext db = new DataContext(Connections.ConnectionString);
            var Orders = db.GetTable<Order>();
            db.Log = Console.Out;

            // 
            // Listing 5-26
            //
            decimal? totalFreight = 
                (from  o in Orders
                 where o.CustomerID == "NOTEXIST"
                 select o).Min( o => (decimal?) o.Freight );
            Console.WriteLine(totalFreight);
        }

        public static void MixedCode() {
            nwDataContext db = new nwDataContext(Connections.ConnectionString);
            db.Log = Console.Out;

            // 
            // Listing 5-30
            //
            var query1 =
                from p in db.Products
                where p.UnitPrice > 50
                select new
                {
                    ProductName = "** " + p.ProductName + " **",
                    p.UnitPrice
                };

            Console.WriteLine(query1.ToString());
            foreach (var row in query1)
            {
                Console.WriteLine(row);
            }

            // 
            // Listing 5-32
            //
            var query2 =
                from p in db.Products
                where p.UnitPrice > 50
                select new
                {
                    ProductName = p.ProductName.Highlight(),
                    p.UnitPrice
                };

            Console.WriteLine(query2.ToString());
            foreach (var row in query2)
            {
                Console.WriteLine(row);
            }

            // 
            // Listing 5-33
            //
            var query3 =
                from a in
                    (
                        from p in db.Products
                        where p.UnitPrice > 50
                        select new
                        {
                            ProductName = p.ProductName.Highlight(),
                            p.UnitPrice
                        }
                        )
                select new
                {
                    ProductName = a.ProductName.ToLower(),
                    a.UnitPrice
                };

            Console.WriteLine(query3.ToString());
            foreach (var row in query3)
            {
                Console.WriteLine(row);
            }

            // 
            // Listing 5-34
            //
            var query4 =
                from a in
                    (
                        from p in db.Products
                        where p.UnitPrice > 50
                        select new
                        {
                            ProductName = "** " + p.ProductName + " **",
                            p.UnitPrice
                        }
                        )
                select new
                {
                    ProductName = a.ProductName.ToLower(),
                    a.UnitPrice
                };

            Console.WriteLine(query4.ToString());
            foreach (var row in query4)
            {
                Console.WriteLine(row);
            }
            
            // 
            // Listing 5-35
            //
            var query5 =
                from p in db.Products
                where p.ProductName.Highlight().Length > 20
                select new {
                    ProductName = p.ProductName.Highlight(),
                    p.UnitPrice
                };

            try {
                Console.WriteLine(query5.ToString());
                foreach (var row in query5) {
                    Console.WriteLine(row);
                }
            }
            catch (NotSupportedException ex) {
                Console.WriteLine(ex.GetType().FullName);
                Console.WriteLine(ex.Message);
            }
        }

    }
    // 
    // Listing 5-31
    //
    static public class Extensions {
        public static string Highlight(this string s) {
            return "** " + s + " **";
        }
    }

    public class CustomerData {
        public CustomerData(string customerID, string companyName) {
            this.CustomerID = customerID;
            this.Name = companyName;
        }
        public CustomerData() {
        }

        public string CustomerID;
        public string Name;
    }

    public class CustomerShortInfo {
        public string CustomerID;
        public string CompanyName;
        public string City;
    }

    public class CustomerInfo {
        public string CustomerID;
        public string CompanyName;
        public string City; 
        public string ContactName;
    }
    public class EmployeeInfo {
        public int EmployeeID;
        public string LastName;
        public string FirstName;
        public int? ReportsTo; // int? Corresponds to Nullable<int>
        public int HierarchyLevel;

        public override string ToString() {
            return String.Format("ID={0} LastName={1} FirstName={2} ReportsTo={3} HierarchyLevel={4}",
                                  this.EmployeeID, this.LastName, this.FirstName, this.ReportsTo, this.HierarchyLevel);
        }

    }

    public class SampleDb : DataContext {
        public SampleDb(IDbConnection connection) : base(connection) { }
        public SampleDb(string fileOrServerOrConnection) : base(fileOrServerOrConnection) { }
        public SampleDb(IDbConnection connection, MappingSource mapping) : base(connection, mapping) { }

        public Table<Customer> Customers;

        // ************************************************
        // ************************************************
        // 
        // Listing 5-12
        // 
        [Function(Name = "Customers by City", IsComposable = false)]
        public ISingleResult<CustomerInfo> CustomersByCity(string param1) {
            IExecuteResult executeResult = 
                this.ExecuteMethodCall(
                         this,
                         (MethodInfo) (MethodInfo.GetCurrentMethod()), // Get Stored Procedure Name
                         param1);    // Stored Procedure Parameter
            ISingleResult<CustomerInfo> result = 
                (ISingleResult<CustomerInfo>) executeResult.ReturnValue;
            return result;                             
        }

        // 
        // Listing 5-13
        // 
        [Function(Name = "TwoCustomerGroups", IsComposable = false)]
        [ResultType(typeof(CustomerInfo))]
        [ResultType(typeof(CustomerShortInfo))]
        public IMultipleResults TwoCustomerGroups() {
            IExecuteResult executeResult = 
                     this.ExecuteMethodCall(
                         this,
                         (MethodInfo) (MethodInfo.GetCurrentMethod()));
            IMultipleResults result =
                (IMultipleResults) executeResult.ReturnValue;
            return result;
        }

        // 
        // Listing 5-14
        // 
        [Function(Name = "dbo.MinUnitPriceByCategory", IsComposable = true)]
        public decimal? MinUnitPriceByCategory( int? categoryID) {
            IExecuteResult executeResult = 
                this.ExecuteMethodCall(
                    this, 
                    ((MethodInfo) (MethodInfo.GetCurrentMethod())), 
                    categoryID);
            decimal? result = (decimal?) executeResult.ReturnValue;
            return result;
        }

        // 
        // Listing 5-15
        // 
        [Function(Name = "dbo.CustomersByCountry", IsComposable = true)]
        public IQueryable<Customer> CustomersByCountry(string country) { 
            return this.CreateMethodCallQuery<Customer>(
                this, 
                ((MethodInfo) (MethodInfo.GetCurrentMethod())), 
                country);
        }
    }
}

// ---------------------------------------------------
//   Stored procedure used by TwoCustomerGroups demo 
// ---------------------------------------------------
    //CREATE PROCEDURE TwoCustomerGroups
    //AS BEGIN
    //    SELECT CustomerID, ContactName, CompanyName, City 
    //    FROM   Customers AS c 
    //    WHERE  c.City = 'London'

    //    SELECT CustomerID, CompanyName, City 
    //    FROM   Customers AS c 
    //    WHERE  c.City = 'Torino'
    //END
// ---------------------------------------------------

// ---------------------------------------------------
//   UDF used by demo 
// ---------------------------------------------------
    //CREATE FUNCTION [dbo].[CustomersByCountry]( @country NVARCHAR(15) )
    //RETURNS TABLE
    //AS
    //RETURN
    //    SELECT CustomerID, ContactName, CompanyName, City
    //    FROM Customers c
    //    WHERE c.Country = @country
// ---------------------------------------------------